@@ -2,10 +2,9 @@
 # this repository contains the full copyright notices and license terms.
 import datetime
 from decimal import Decimal
-from itertools import groupby, ifilter
+from itertools import groupby
 from functools import partial
 
-from trytond import backend
 from trytond.model import ModelView, ModelSQL, fields
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 from trytond.pyson import If, In, Eval, Bool
@@ -31,11 +30,7 @@ class PurchaseRequest(ModelSQL, ModelView):
     product = fields.Many2One('product.product', 'Product',
         select=True, readonly=True, domain=[('purchasable', '=', True)])
     description = fields.Text('Description', readonly=True,
-        states={
-            'required': ~Eval('product'),
-            'readonly': STATES['readonly'],
-            },
-        depends=['product'] + DEPENDS)
+        states=STATES, depends=DEPENDS)
     party = fields.Many2One('party.party', 'Party', select=True, states=STATES,
         depends=DEPENDS)
     quantity = fields.Float('Quantity', required=True, states=STATES,
@@ -99,6 +94,7 @@ class PurchaseRequest(ModelSQL, ModelView):
         cls._buttons.update({
                 'handle_purchase_cancellation_exception': {
                     'invisible': Eval('state') != 'exception',
+                    'depends': ['state'],
                     },
                 })
 
@@ -108,19 +104,18 @@ class PurchaseRequest(ModelSQL, ModelView):
         ModelData = pool.get('ir.model.data')
         Purchase = pool.get('purchase.purchase')
         PurchaseLine = pool.get('purchase.line')
-        TableHandler = backend.get('TableHandler')
         model_data = ModelData.__table__()
         purchase = Purchase.__table__()
         purchase_line = PurchaseLine.__table__()
         request = cls.__table__()
 
-        tablehandler = TableHandler(cls, module_name)
+        tablehandler = cls.__table_handler__(module_name)
         state_exist = tablehandler.column_exist('state')
 
         super(PurchaseRequest, cls).__register__(module_name)
 
         # Migration from 3.6: removing the constraint on the quantity
-        tablehandler = TableHandler(cls, module_name)
+        tablehandler = cls.__table_handler__(module_name)
         tablehandler.drop_constraint('check_purchase_request_quantity')
 
         # Migration from 3.8: renaming module of Purchase Request group entry
@@ -315,7 +310,7 @@ class CreatePurchase(Wizard):
     ask_party = StateView('purchase.request.create_purchase.ask_party',
         'purchase_request.purchase_request_create_purchase_ask_party', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Continue', 'start', 'tryton-go-next', default=True),
+            Button('Continue', 'start', 'tryton-forward', default=True),
             ])
 
     @classmethod
@@ -387,7 +382,7 @@ class CreatePurchase(Wizard):
                     and request.product == self.ask_party.product
                     and compare_string(
                         request.description, self.ask_party.description))
-            reqs = filter(to_write, requests)
+            reqs = list(filter(to_write, requests))
             if reqs:
                 Request.write(reqs, {
                         'party': self.ask_party.party.id,
@@ -399,7 +394,7 @@ class CreatePurchase(Wizard):
 
         def to_ask_party(request):
             return not request.purchase_line and not request.party
-        reqs = ifilter(to_ask_party, requests)
+        reqs = filter(to_ask_party, requests)
         if any(reqs):
             return 'ask_party'
 
@@ -439,55 +434,19 @@ class CreatePurchase(Wizard):
         Request.update_state(requests)
         return 'end'
 
-    @staticmethod
-    def _get_tax_rule_pattern(line, purchase):
-        '''
-        Get tax rule pattern
-        '''
-        return {}
-
     @classmethod
     def compute_purchase_line(cls, key, requests, purchase):
         pool = Pool()
-        Product = pool.get('product.product')
         Line = pool.get('purchase.line')
 
         line = Line()
+        line.unit_price = Decimal(0).quantize(
+            Decimal(1) / 10 ** Line.unit_price.digits[1])
         for f, v in key:
             setattr(line, f, v)
-        if not line.description:
-            line.description = line.product.name
         line.quantity = sum(r.quantity for r in requests)
-
-        if not getattr(line, 'unit_price', None):
-            if line.product and line.unit:
-                with Transaction().set_context(uom=line.unit.id,
-                        supplier=purchase.party.id,
-                        currency=purchase.currency.id):
-                    product_price = Product.get_purchase_price(
-                        [line.product], line.quantity)[line.product.id]
-            else:
-                product_price = Decimal(0)
-            product_price = product_price.quantize(
-                Decimal(1) / 10 ** Line.unit_price.digits[1])
-            line.unit_price = product_price
-
-        taxes = []
-        pattern = cls._get_tax_rule_pattern(line, purchase)
-        if line.product:
-            for tax in line.product.supplier_taxes_used:
-                if purchase.party and purchase.party.supplier_tax_rule:
-                    tax_ids = purchase.party.supplier_tax_rule.apply(
-                        tax, pattern)
-                    if tax_ids:
-                        taxes.extend(tax_ids)
-                    continue
-                taxes.append(tax.id)
-        if purchase.party and purchase.party.supplier_tax_rule:
-            tax_ids = purchase.party.supplier_tax_rule.apply(None, pattern)
-            if tax_ids:
-                taxes.extend(tax_ids)
-            line.taxes = taxes
+        line.purchase = purchase
+        line.on_change_product()
         return line
 
 
